// This file handles the nitty-gritty hardware details.

#include <avr/io.h>

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

// Pin constants
const int RCLK = 2, SRCLK = 3,  SER1 = 4, SER2 = 9,
          BUTTON1 = 5, BUTTON2 = 6, BUTTON3 = 7, BUTTON4 = 8,
          LED1 = 10, LED2 = 11, LED3 = 12, LED4 = 13;

// Handled automatically by the interrupt.
char LEDs[4][4];
boolean buttons[4][4];

void initHardware() {
  // Set all relevant pins at outputs
  const int outputs[] = {
    RCLK, SRCLK, SER1, SER2, LED1, LED2, LED3, LED4};

  // Make everything outputs
  for (int i = 0; i < 8; ++i)
    pinMode(outputs[i], OUTPUT);
  
  // Initialize LED and buttons to all off
  for (int i = 0; i < 4; ++i)
    for (int j = 0; j < 4; ++j) {
      buttons[i][j] = 0;
      LEDs[i][j] = 0;
    }
  
  // Timer 2 mode = normal
  TCCR2A = 0;
  // Timer 2 prescalar = 32
  TCCR2B = 0b011;  
  sbi(TIMSK2, TOIE2); // Enable Timer2 interrupts
}

void setColor(int i, int j, char color)
{
  LEDs[i][j] = color;
}

char getColor(int i, int j)
{
  return LEDs[i][j];
}

boolean getButton(int i, int j)
{
  return buttons[i][j];
}

// This interrupt handles the system hardware.  It triggers about every 0.5 ms.
ISR(TIMER2_OVF_vect) {
  // Switch between columns every time this routine is called.
  static int column = 3;
  // Switch between rows of active button every time this is called.
  static int button = 1;
  // Each column has four pulses.
  static int pulse = 0;
  
  // Output register clock goes low...
  digitalWrite(RCLK, LOW);

  static int COLUMN_PINS[] = {LED4, LED1, LED2, LED3};

  for (int i = 0; i < 8; ++i)
  {
    // Shift register clock goes low
    digitalWrite(SRCLK, LOW);

    // Write out data
    if (i == 0) {
      digitalWrite(SER1, (LEDs[1][column] & 0b11) > pulse);
      digitalWrite(SER2, (LEDs[3][column] & 0b11) > pulse);
    }
    else if (i == 1) {
      digitalWrite(SER1, (LEDs[1][column] & 0b1100) > (pulse << 2));
      digitalWrite(SER2, (LEDs[3][column] & 0b1100) > (pulse << 2));
    }
    else if (i == 2) {
      digitalWrite(SER1, (LEDs[1][column] & 0b110000) > (pulse << 4));
      digitalWrite(SER2, (LEDs[3][column] & 0b110000) > (pulse << 4));
    }
    else if (i == 3) { // Buttons 2 and 4
      digitalWrite(SER1, button == 2);
      digitalWrite(SER2, button == 4);
    } 
    else if (i == 4) {
      digitalWrite(SER1, (LEDs[0][column] & 0b11) > pulse);
      digitalWrite(SER2, (LEDs[2][column] & 0b11) > pulse);
    }
    else if (i == 5) {
      digitalWrite(SER1, (LEDs[0][column] & 0b1100) > (pulse << 2));
      digitalWrite(SER2, (LEDs[2][column] & 0b1100) > (pulse << 2));
    }
    else if (i == 6) {
      digitalWrite(SER1, (LEDs[0][column] & 0b110000) > (pulse << 4));
      digitalWrite(SER2, (LEDs[2][column] & 0b110000) > (pulse << 4));
    }
    else if (i == 7) { // Buttons 1 and 3
      digitalWrite(SER1, button == 1);
      digitalWrite(SER2, button == 3);
    }

    // Shift register clock goes high!
    digitalWrite(SRCLK, HIGH);
  }

  // Turn off the previous column if we're at a new pulse
  if (pulse == 0)
    digitalWrite(COLUMN_PINS[column], LOW);

  // Output register clock goes high!  
  digitalWrite(RCLK, HIGH);

  // And finally, light up the night (if we're at a new pulse)
  if (pulse == 0) {
    digitalWrite(COLUMN_PINS[(column + 1) % 4], HIGH);  
  }
  
  // If we're about to move to the next column, update various things.
  if (++pulse == 4) {
    pulse = 0;
      
    // And now save new buttons.
    buttons[button-1][0] = digitalRead(BUTTON1);
    buttons[button-1][1] = digitalRead(BUTTON2);
    buttons[button-1][2] = digitalRead(BUTTON3);
    buttons[button-1][3] = digitalRead(BUTTON4);
    
    // And finally, shift over so that we're looking at the next row.
    if (++button == 5)
        button = 1;
    
    if (++column == 4)
      column = 0;
  }
}
